#include <asm/system.h>
#include <asm/io.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sched.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <linux/mc146818rtc.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <time.h>

#define RTC_FREQ 1024
#define MSEC_DELAY 25

int ports[] = {0x378, 0x278, 0x3BC, 0};
int write_to_rtc = 0;

typedef struct
 {
  unsigned char flags;
  unsigned char sec;
  unsigned char min;
  unsigned char hour;
  unsigned char wday;
  unsigned char mday;
  unsigned char month;
  unsigned char year;
  unsigned char flags2;
 } dcf77;

int rtc_fd;
int rtc_rate;

void msleep_init()
 {
  int r;
  
  rtc_fd = open("/dev/rtc", O_RDONLY);
  if (rtc_fd == -1)
   {
    fprintf(stderr, "Could not open /dev/rtc. Support compiled into kernel?\n");
    exit(1);
   }
   
  r = ioctl(rtc_fd, RTC_IRQP_READ, &rtc_rate); 
  if (r == -1)
   {
    fprintf(stderr, "Could not read interrupt rate.\n");
    exit (2);
   }

  r = ioctl(rtc_fd, RTC_IRQP_SET, RTC_FREQ);
  if (r == -1)
   {
    fprintf(stderr, "Could not set interrupt rate.\n");
    exit(3);
   }
   
  r = ioctl(rtc_fd, RTC_PIE_ON, 0); 
  if (r == -1)
   {
    fprintf(stderr, "Could not enable interrupts.\n");
    exit(5);
   }
 }

void msleep_done()
 {
  int r;
  r = ioctl(rtc_fd, RTC_PIE_OFF, 0); 
  if (r == -1)
   {
    fprintf(stderr, "Could not disable interrupts.\n");
    exit(5);
   }

  r = ioctl(rtc_fd, RTC_IRQP_SET, rtc_rate); 
  if (r == -1)
   {
    fprintf(stderr, "Could not reset interrupt rate.\n");
    exit(4);
   }

  close(rtc_fd);
 }

void msleep(int delay)
 {
  int r, v;
  unsigned long data;

  delay = (100*delay)/RTC_FREQ;
  
  r = read(rtc_fd, &data, sizeof(data));
  if (r == -1) 
   {
    fprintf(stderr, "Could not read value from rtc.\n");
    exit(6);
   }
  
  v = delay;  
  while (v > 0)
   {
    r = read(rtc_fd, &data, sizeof(data));
    if (r == -1) 
     {
      fprintf(stderr, "Could not read value from rtc.\n");
      exit(6);
     }
     
    v -= (data >> 8);
   }
 }

#define bsdtoint(x) ((x>>4) + ((x&15)*10))

int calc_time(dcf77 *d)
 {
  struct tm t;
  time_t oldt, newt;
  
  if (d->flags != 9)
   {
    t.tm_sec   = bsdtoint(d->sec);
    t.tm_min   = bsdtoint(d->min);  
    t.tm_hour  = bsdtoint(d->hour);  
    t.tm_mday  = bsdtoint(d->mday);
    t.tm_mon   = bsdtoint(d->month)-1;  
    t.tm_year  = bsdtoint(d->year);  
    t.tm_wday  = (d->wday & 7) == 7 ? 0 : d->wday & 7;  
    t.tm_isdst = (d->wday & 16) ? 1 : 0;

    if (d->wday & 64) 
     printf("dcf77: !! LOW BATT !!\n");

    oldt = time(0);
    newt = mktime(&t);
   
    printf("dcf77: DCF: %s", asctime (&t));
    printf("dcf77: RTC: %s", ctime (&oldt));


    printf("dcf77: Time difference is %i second(s).\n", newt-oldt);
   
    if (write_to_rtc) 
     {
      struct rtc_time rtct;
      int r;
      
      rtct.tm_sec = t.tm_sec;
      rtct.tm_min = t.tm_min;
      rtct.tm_hour = t.tm_hour;
      rtct.tm_mday = t.tm_mday;
      rtct.tm_mon = t.tm_mon;
      rtct.tm_year = t.tm_year;
      rtct.tm_wday = t.tm_wday;
      rtct.tm_yday = t.tm_yday;
      rtct.tm_isdst = t.tm_isdst;
      
      r = ioctl(rtc_fd, RTC_SET_TIME, &rtct);
      if (r == -1)
       {
        fprintf(stderr, "Could not write to RTC.\n");
        exit(9);
       }

      printf("dcf77: Time written to RTC.\n"); 
      
      stime(&newt);
      printf("dcf77: Time written to system.\n"); 
     }
   }    
  else
   printf("dcf77: Time corrupted.\n");
 
 }

int read_time(int port, dcf77 *d, int tm)
 {
  int bit, i, c, lc, timeout;
  unsigned char *b, *p;
  struct sched_param sched;
  
  sched.sched_priority = 1;
  sched_setscheduler(0, SCHED_FIFO, &sched); 
  
  for (i = 0; i <= 7; i++) 
   {
    outb(i & 7, port);
    outb(i & 1 ? 4 : 6, port+2);
    msleep(40);
   }

  outb(0, port);
  outb(6, port+2);
  msleep(40);
  outb(4, port+2);

  memset(d, 0, sizeof(dcf77));
  b = (unsigned char*) d;

  lc = (inb(port+1) & 16) ? 1 : 0;

  for (bit = 0; bit <= 68; bit ++)
   {
    *b <<= 1;

    timeout = 50000;
    do
     {
      c = (inb(port+1) & 16) ? 1 : 0; 
      timeout--;
      if (timeout <= 0) return 0;
     } 
    while (c == lc);
    lc = c;

    *b |= (inb(port+1) >> 5) & 1;
    
    if ((bit & 7) == 7) 
     b++;   
   }
   
  return 1; 
 }   

int try_port(int p, int tm) 
 {
  dcf77 d;
  if (read_time(p, &d, tm)) 
   {
    calc_time(&d);
    return 1;
   }
  else
   return 0;     
 }

int main(int argc, char *argv[])
 {
  int i, p;
  if (iopl(3)) 
   {
    fprintf(stderr, "Must be root.\n");
    exit(8);
   }
   
  for (i = 1; i < argc; i++)
   {
    if (strcmp(argv[i], "-s") == 0)
     {
      write_to_rtc = 1;
     } 
    else if (strncmp(argv[i], "0x", 2) == 0)
     {
      ports[0] = strtoul(argv[i], 0, 16);
      ports[1] = 0;
     }
    else
     {
      fprintf(stderr, "Unknown command.\n");
      printf("Syntax: %s [-s] [port]\n", argv[0]);
      printf("        -s : Write to RTC, otherwise time is just printed to console.\n");
      printf("      port : You may specify the base port of the parellel line in hexadecimal form 0xFFF.\n");
      printf("             0x378 => /dev/lp0; 0x278 => /dev/lp1; 0x3BC => /dev/lp2.\n");
      printf("             If not given, %s tries to make an automatic check for it. (plug'n'play).", argv[0]);
           exit(9);
     }
   } 
  
  msleep_init();
  
  i = 0;
  while (p = ports[i++])
   {
    printf("dcf77: Trying port 0x%x.\n", p);
    if (try_port(p, 25)) break;
    if (try_port(p, 25)) break;

    if (try_port(p, 40)) break;
    if (try_port(p, 40)) break;

    if (try_port(p, 30)) break;
    if (try_port(p, 30)) break;

    if (try_port(p, 35)) break;
    if (try_port(p, 35)) break;
   }      

  if (p == 0)
   {
    fprintf(stderr, "Could not detect clock.\n");
    exit(7);
   } 
   
  msleep_done();    
  return 0;
 }